/* ***** BEGIN LICENSE BLOCK *****
 * Version: GPL 2.0
 *
 * The contents of this file are subject to the GNU General Public
 * License Version 2 or later (the "GPL").
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Initial Developer of the Original Code is
 *   MiniG.org project members
 *
 * ***** END LICENSE BLOCK ***** */

package org.obm.push.backend.obm22.mail;

import java.io.BufferedInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.sql.SQLException;
import java.util.Arrays;
import java.util.Collection;
import java.util.LinkedList;
import java.util.List;
import java.util.Set;
import java.util.concurrent.ConcurrentMap;
import java.util.concurrent.TimeUnit;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.columba.ristretto.message.Address;
import org.minig.imap.FastFetch;
import org.minig.imap.Flag;
import org.minig.imap.FlagsList;
import org.minig.imap.IMAPException;
import org.minig.imap.ListInfo;
import org.minig.imap.ListResult;
import org.minig.imap.SearchQuery;
import org.minig.imap.StoreClient;
import org.minig.obm.pool.OBMPoolActivator;
import org.obm.locator.client.LocatorClient;
import org.obm.push.backend.BackendSession;
import org.obm.push.backend.EmailCache;
import org.obm.push.backend.FilterType;
import org.obm.push.backend.MSEmail;
import org.obm.push.backend.obm22.mail.smtp.SmtpLocator;
import org.obm.push.backend.obm22.mail.smtp.SmtpSender;
import org.obm.push.exception.ProcessingEmailException;
import org.obm.push.exception.SendEmailException;
import org.obm.push.exception.ServerErrorException;
import org.obm.push.exception.SmtpInvalidRcptException;
import org.obm.push.exception.StoreEmailException;
import org.obm.push.state.SyncState;
import org.obm.push.store.ISyncStorage;
import org.obm.sync.client.calendar.CalendarClient;

import com.google.common.base.Function;
import com.google.common.collect.Collections2;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.MapMaker;

import fr.aliasource.utils.IniFile;

/**
 * 
 * @author adrienp
 * 
 */
public class EmailManager implements IEmailManager {

	public static final String IMAP_INBOX_NAME = "INBOX";
	private static final String BACKEND_CONF_FILE = "/etc/opush/mail_conf.ini";
	private static final String BACKEND_IMAP_LOGIN_WITH_DOMAIN = "imap.loginWithDomain";
	private static final String BACKEND_IMAP_ACTIVATE_TLS = "imap.activateTLS";

	private ISyncStorage syncStore;
	private SmtpSender smtpProvider;

	private Boolean loginWithDomain;
	private Boolean activateTLS;
	protected String imapHost;

	private Log logger;
	private ConcurrentMap<Integer, IEmailSync> uidCache;

	public EmailManager(ISyncStorage syncStore) {
		this.logger = LogFactory.getLog(getClass());
		this.smtpProvider = new SmtpSender(new SmtpLocator()); 
		this.syncStore = syncStore;
		this.uidCache = initUidCacheMap();
		IniFile ini = new IniFile(BACKEND_CONF_FILE) {
			@Override
			public String getCategory() {
				return null;
			}
		};
		loginWithDomain = !"false".equals(ini.getData().get(
				BACKEND_IMAP_LOGIN_WITH_DOMAIN));
		activateTLS = !"false".equals(ini.getData().get(
				BACKEND_IMAP_ACTIVATE_TLS));
	}

	private ConcurrentMap<Integer, IEmailSync> initUidCacheMap() {
		return new MapMaker()
	       .concurrencyLevel(16)
	       .expiration(1, TimeUnit.DAYS)
	       .weakValues()
	       .makeMap();
	}


	private void locateImap(BackendSession bs) {
		imapHost = new LocatorClient().locateHost("mail/imap_frontend",
				bs.getLoginAtDomain());
		logger.info("Using " + imapHost + " as imap host.");
	}

	private StoreClient getImapClient(BackendSession bs) {
		if (imapHost == null) {
			locateImap(bs);
		}
		String login = bs.getLoginAtDomain();
		if (!loginWithDomain) {
			int at = login.indexOf("@");
			if (at > 0) {
				login = login.substring(0, at);
			}
		}
		logger.info("creating storeClient with login: " + login
				+ " (loginWithDomain: " + loginWithDomain + ", activateTLS:"
				+ activateTLS + ")");
		StoreClient imapCli = new StoreClient(imapHost, 143, login,
				bs.getPassword());

		return imapCli;
	}

	

	private IEmailSync cache(Integer collectionId, Boolean reset) {
		IEmailSync ret = uidCache.get(collectionId);
		if (ret == null) {
			ret = new EmailSyncCache(syncStore, OBMPoolActivator.getDefault());
			uidCache.put(collectionId, ret);
		}
		return ret;
	}

	@Override
	public MailChanges getSync(BackendSession bs, SyncState state,
			Integer devId, Integer collectionId, String collectionName,
			FilterType filter) throws ServerErrorException {
		IEmailSync uc = cache(collectionId, false);
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBox = parseMailBoxName(store, bs, collectionName);
			store.select(mailBox);
			MailChanges sync = uc.getSync(store, devId, bs, state,
					collectionId, filter);
			return sync;
		} catch (Throwable e) {
			throw new ServerErrorException(e);
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
				logger.error(e);
			}
		}
	}

	@Override
	public List<MSEmail> fetchMails(BackendSession bs, CalendarClient calendarClient, Integer collectionId,
			String collectionName, Set<Long> uids) throws IOException, IMAPException {	
		final List<MSEmail> mails = new LinkedList<MSEmail>();
		final StoreClient store = getImapClient(bs);
		try {
			
			login(store);
			store.select(parseMailBoxName(store, bs, collectionName));
			
			final MailMessageLoader mailLoader = new MailMessageLoader(store, calendarClient);
			for (final Long uid: uids) {
				final MSEmail email = mailLoader.fetch(collectionId, uid, bs);
				if (email != null) {
					mails.add(email);
				}
			}
			
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
				logger.error(e);
			}
		}
		return mails;
	}

	private ListResult listAllSubscribedFolder(StoreClient store,
			BackendSession bs) throws IMAPException {
		try {
			return store.listSubscribed("", "*");
		} catch (Throwable e) {
			logger.error("Can't read list of subscribe folder.", e);
			throw new IMAPException("Can't read list of subscribe folder");
		}

	}

	@Override
	public void updateReadFlag(BackendSession bs, String collectionName,
			Long uid, boolean read) throws IMAPException {
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBoxName = parseMailBoxName(store, bs, collectionName);
			store.select(mailBoxName);
			FlagsList fl = new FlagsList();
			fl.add(Flag.SEEN);
			store.uidStore(Arrays.asList(uid), fl, read);
			logger.info("flag  change: " + (read ? "+" : "-") + " SEEN"
					+ " on mail " + uid + " in " + mailBoxName);
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
	}

	@Override
	public String parseMailBoxName(BackendSession bs, String collectionName)
			throws IMAPException {
		// parse obm:\\adrien@test.tlse.lng\email\INBOX\Sent
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			return parseMailBoxName(store, bs, collectionName);
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
	}

	private String parseMailBoxName(StoreClient store, BackendSession bs,
			String collectionName) throws IMAPException {
		// parse obm:\\adrien@test.tlse.lng\email\INBOX\
		if (collectionName.toLowerCase()
				.endsWith(IMAP_INBOX_NAME.toLowerCase())) {
			return IMAP_INBOX_NAME;
		}
		// parse obm:\\adrien@test.tlse.lng\email\Sent
		int slash = collectionName.lastIndexOf("email\\");
		String boxName = collectionName.substring(slash + "email\\".length());
		ListResult lr = listAllSubscribedFolder(store, bs);
		for (ListInfo i : lr) {
			if (i.getName().toLowerCase().contains(boxName.toLowerCase())) {
				return i.getName();
			}
		}
		throw new IMAPException("Cannot find IMAP folder for collection["
				+ collectionName + "]");
	}

	@Override
	public void resetForFullSync(Set<Integer> listCollectionId) {
		for (Integer colId : listCollectionId) {
			uidCache.remove(colId);
		}
	}

	@Override
	public void delete(BackendSession bs, Integer devId, String collectionPath,
			Integer collectionId, Long uid) throws IMAPException {
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBoxName = parseMailBoxName(store, bs, collectionPath);
			store.select(mailBoxName);
			FlagsList fl = new FlagsList();
			fl.add(Flag.DELETED);
			logger.info("delete conv id : " + uid);
			store.uidStore(Arrays.asList(uid), fl, true);
			store.expunge();
			deleteMessageInCache(devId, collectionId, Arrays.asList(uid));
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
	}

	@Override
	public Long moveItem(BackendSession bs, Integer devId, String srcFolder,
			Integer srcFolderId, String dstFolder, Integer dstFolderId, Long uid)
			throws IMAPException {
		StoreClient store = getImapClient(bs);
		Collection<Long> newUid = null;
		try {
			login(store);
			String srcMailBox = parseMailBoxName(store, bs, srcFolder);
			String dstMailBox = parseMailBoxName(store, bs, dstFolder);
			store.select(srcMailBox);
			List<Long> uids = Arrays.asList(uid);
			newUid = store.uidCopy(uids, dstMailBox);
			FlagsList fl = new FlagsList();
			fl.add(Flag.DELETED);
			logger.info("delete conv id : " + uid);
			store.uidStore(uids, fl, true);
			store.expunge();
			deleteMessageInCache(devId, srcFolderId, Arrays.asList(uid));
			addMessageInCache(store, devId, dstFolderId, uid);
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
		if (newUid == null || newUid.isEmpty()) {
			return null;
		}
		return newUid.iterator().next();
	}

	@Override
	public List<InputStream> fetchMIMEMails(BackendSession bs,
			CalendarClient calendarClient, String collectionName, Set<Long> uids)
			throws IOException, IMAPException {
		List<InputStream> mails = new LinkedList<InputStream>();
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			store.select(parseMailBoxName(store, bs, collectionName));
			for (Long uid : uids) {
				mails.add(store.uidFetchMessage(uid));
			}
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
		return mails;
	}

	private void login(StoreClient store) throws IMAPException {
		if (!store.login(activateTLS)) {
			throw new IMAPException("Cannot log into imap server");
		}
	}

	@Override
	public void setAnsweredFlag(BackendSession bs, String collectionName,
			Long uid) throws IMAPException {
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBoxName = parseMailBoxName(store, bs, collectionName);
			store.select(mailBoxName);
			FlagsList fl = new FlagsList();
			fl.add(Flag.ANSWERED);
			store.uidStore(Arrays.asList(uid), fl, true);
			logger.info("flag  change: "
					+ ("+ ANSWERED" + " on mail " + uid + " in " + mailBoxName));
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
	}

	@Override
	public void sendEmail(BackendSession bs, Address from, Set<Address> setTo,
			Set<Address> setCc, Set<Address> setCci, InputStream mimeMail,
			Boolean saveInSent) throws ProcessingEmailException, SendEmailException, SmtpInvalidRcptException {
		SmtpInvalidRcptException invalidRctp = null;
		InputStream streamMail = new BufferedInputStream(mimeMail);
		try {
			try{
				streamMail.mark(streamMail.available());
			} catch (IOException e) {
				throw new ProcessingEmailException();
			} 
			
			try {
				smtpProvider.sendEmail(bs, from, setTo, setCc, setCci, streamMail);
			} catch (SmtpInvalidRcptException e1) {
				invalidRctp = e1;
			}
			
			if (saveInSent) {
				try {
					streamMail.reset();
					storeInSent(bs, streamMail);
				} catch (Throwable e) {
					logger.error("Error during store mail in Sent folder", e);
				}
			}
		} finally {
			closeStream(streamMail);
		}
		if(invalidRctp != null){
			throw invalidRctp;
		}
	}	
	
	private void closeStream(InputStream mimeMail) {
		if (mimeMail != null) {
			try {
				mimeMail.close();
			} catch (Throwable t) {
			}
		}
	}
	
	@Override
	public InputStream findAttachment(BackendSession bs, String collectionName,
			Long mailUid, String mimePartAddress) throws IMAPException {

		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBoxName = parseMailBoxName(store, bs, collectionName);
			store.select(mailBoxName);
			return store.uidFetchPart(mailUid, mimePartAddress);
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}
	}

	@Override
	public void purgeFolder(BackendSession bs, Integer devId,
			String collectionPath, Integer collectionId) throws IMAPException {
		long time = System.currentTimeMillis();
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String mailBoxName = parseMailBoxName(store, bs, collectionPath);
			store.select(mailBoxName);
			logger.info("Mailbox folder[" + collectionPath
					+ "] will be purged...");
			Collection<Long> uids = store.uidSearch(new SearchQuery());
			FlagsList fl = new FlagsList();
			fl.add(Flag.DELETED);
			store.uidStore(uids, fl, true);
			store.expunge();
			deleteMessageInCache(devId, collectionId, uids);
			time = System.currentTimeMillis() - time;
			logger.info("Mailbox folder[" + collectionPath + "] was purged in "
					+ time + " millisec. " + uids.size()
					+ " messages have been deleted");
		} finally {
			try {
				store.logout();
			} catch (IMAPException e) {
			}
		}

	}

	@Override
	public Long storeInInbox(BackendSession bs, InputStream mailContent, boolean isRead)
			throws StoreEmailException {
		logger.info("Store mail in folder[Inbox]");
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			return storeMail(store, IMAP_INBOX_NAME, isRead, mailContent, false);
		} catch (Throwable e) {
			throw new StoreEmailException(
					"Error during store mail in Inbox folder", e);
		} finally {
			try {
				store.logout();
			} catch (Throwable e) {
			}
		}
	}

	/**
	 * Store the mail in the Sent folder storeInSent reset the mimeMail will be
	 * if storeInSent read it
	 * 
	 * @param bs
	 *            the BackendSession
	 * @param mail
	 *            the mail that will be stored
	 * @return the imap uid of the mail
	 * @throws StoreEmailException
	 */
	private long storeInSent(BackendSession bs, InputStream mail)
			throws StoreEmailException {
		logger.info("Store mail in folder[Sent]");
		StoreClient store = getImapClient(bs);
		try {
			login(store);
			String sentFolderName = null;
			ListResult lr = listAllSubscribedFolder(store, bs);
			for (ListInfo i : lr) {
				if (i.getName().toLowerCase().endsWith("sent")) {
					sentFolderName = i.getName();
				}
			}
			return storeMail(store, sentFolderName,true, mail, true);
		} catch (Throwable e) {
			throw new StoreEmailException(
					"Error during store mail in Sent folder", e);
		} finally {
			try {
				store.logout();
			} catch (Throwable e) {
			}
		}
	}

	/**
	 * 
	 * @param store
	 *            the StoreClient
	 * @param folderName
	 *            the folder name where the mail will be stored
	 * @param isRead
	 *            if true the message will be stored with SEEN Flag
	 * @param reset
	 *            if true mailContent will be reseted
	 * @return the imap uid of the mail
	 */
	private Long storeMail(StoreClient store, String folderName,
			boolean isRead, InputStream mailContent, boolean reset) {
		Long ret = null;
		if (folderName != null) {
			if (reset && mailContent.markSupported()) {
				mailContent.mark(0);
			}
			FlagsList fl = new FlagsList();
			if(isRead){
				fl.add(Flag.SEEN);
			}
			ret = store.append(folderName, mailContent, fl);
			store.expunge();
		}
		return ret;
	}

	private void deleteMessageInCache(Integer devId, Integer collectionId,
			Collection<Long> mailUids) {
		try {
			syncStore.removeMessages(devId, collectionId, mailUids);
		} catch (SQLException e) {
			logger.error("Error while deleting messages in db", e);
		}
	}

	private void addMessageInCache(StoreClient store, Integer devId,
			Integer collectionId, Long mailUids) {
		Collection<FastFetch> fetch = store.uidFetchFast(ImmutableSet
				.of(mailUids));
		Collection<EmailCache> emails = Collections2.transform(fetch,
				new Function<FastFetch, EmailCache>() {
					@Override
					public EmailCache apply(FastFetch input) {
						return new EmailCache(input.getUid(), input.isRead());
					}
				});
		try {
			syncStore.addMessages(devId, collectionId, emails);
		} catch (SQLException e) {
			logger.error("Error while adding messages in db", e);
		}
	}

	@Override
	public Boolean getLoginWithDomain() {
		return loginWithDomain;
	}

	@Override
	public Boolean getActivateTLS() {
		return activateTLS;
	}

}
