package org.obm.push.backend.obm22.calendar;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.obm.push.backend.BackendSession;
import org.obm.push.backend.IApplicationData;
import org.obm.push.backend.MSAttendee;
import org.obm.push.backend.MSEvent;
import org.obm.push.data.calendarenum.AttendeeStatus;
import org.obm.push.data.calendarenum.AttendeeType;
import org.obm.sync.calendar.Attendee;
import org.obm.sync.calendar.Event;
import org.obm.sync.calendar.EventRecurrence;
import org.obm.sync.calendar.ParticipationRole;
import org.obm.sync.calendar.ParticipationState;
import org.obm.sync.calendar.RecurrenceKind;

public class EventConverterTest {
	
	private EventConverter converter;
	
	@Before
	public void prepareEventConverter(){
		converter = new EventConverter();
	}
	
	private BackendSession getFakeBackendSession(){
		return new BackendSession("login@domain.fr", "test", "devId", "devType", "Sync");
	}
	
	private Event getFakeEvent(){
		Event e = new Event();
		e.setDate(new Date());
		e.setRecurrence(new EventRecurrence());
		e.getRecurrence().setKind(RecurrenceKind.none);
		return e;
	}
	
	private MSEvent getFakeMSEvent(){
		MSEvent e = new MSEvent();
		e.setStartTime(new Date());
		e.setEndTime(new Date());
		e.setSubject("Subject");
		return e;
	}
	
	@Test
	public void testConvertEventWithOrganizer(){
		Event e = getFakeEvent();
		Attendee att1 = new Attendee();
		att1.setDisplayName("Login Test");
		att1.setEmail("login@domain.fr");
		att1.setObmUser(true);
		att1.setRequired(ParticipationRole.REQ);
		att1.setState(ParticipationState.ACCEPTED);
		e.addAttendee(att1);
		
		Attendee att2 = new Attendee();
		att2.setDisplayName("User Test");
		att2.setEmail("userTest@domain.fr");
		att2.setObmUser(true);
		att2.setOrganizer(true);
		att2.setRequired(ParticipationRole.REQ);
		att2.setState(ParticipationState.ACCEPTED);
		e.addAttendee(att2);
		
		IApplicationData ad = converter.convert(getFakeBackendSession(), e);
		Assert.assertTrue(ad instanceof MSEvent);
		MSEvent mse = (MSEvent) ad;
		Assert.assertEquals("userTest@domain.fr", mse.getOrganizerEmail());
		Assert.assertEquals("User Test", mse.getOrganizerName());
		Assert.assertEquals(2, mse.getAttendees().size());
		
	}
	
	@Test
	public void testConvertEventWithoutOrganizer(){
		Event e = getFakeEvent();
		Attendee att1 = new Attendee();
		att1.setDisplayName("Login Test");
		att1.setEmail("login@domain.fr");
		att1.setObmUser(false);
		att1.setRequired(ParticipationRole.REQ);
		att1.setState(ParticipationState.ACCEPTED);
		e.addAttendee(att1);
		
		Attendee att2 = new Attendee();
		att2.setDisplayName("User Test");
		att2.setEmail("userTest@domain.fr");
		att2.setObmUser(true);
		att2.setRequired(ParticipationRole.REQ);
		att2.setState(ParticipationState.ACCEPTED);
		e.addAttendee(att2);
		
		IApplicationData ad = converter.convert(getFakeBackendSession(), e);
		Assert.assertTrue(ad instanceof MSEvent);
		MSEvent mse = (MSEvent) ad;
		Assert.assertEquals("login@domain.fr", mse.getOrganizerEmail());
		Assert.assertEquals("Login Test", mse.getOrganizerName());
		Assert.assertEquals(2, mse.getAttendees().size());
	}
	
	@Test
	public void testConvertMsEventWithoutOrganizerAndAttendee(){
		MSEvent e = new MSEvent();
		e.setStartTime(new Date());
		e.setEndTime(new Date());
		e.setSubject("Subject");
		BackendSession bs = getFakeBackendSession();
		Event obmEvent = converter.convertAsExternal(bs, e);
		
		Assert.assertEquals(1, obmEvent.getAttendees().size());
		Attendee att = obmEvent.getAttendees().iterator().next();
		Assert.assertEquals(bs.getLoginAtDomain(), att.getEmail());
		Assert.assertEquals(ParticipationRole.REQ, att.getRequired());
		Assert.assertEquals(ParticipationState.ACCEPTED, att.getState());
		Assert.assertTrue(att.isOrganizer());
	}
	
	@Test
	public void testConvertAttendee(){
		MSEvent e = new MSEvent();
		e.setStartTime(new Date());
		e.setEndTime(new Date());
		e.setSubject("Subject");
		MSAttendee newAttendee = new MSAttendee();
		newAttendee.setName("displayName");
		newAttendee.setEmail("email@domain.com");
		newAttendee.setAttendeeType(AttendeeType.REQUIRED);
		newAttendee.setAttendeeStatus(AttendeeStatus.ACCEPT);
		e.addAttendee(newAttendee);
		
		BackendSession bs = getFakeBackendSession();
		Event obmEvent = converter.convertAsExternal(bs, e);
		Assert.assertEquals(2, obmEvent.getAttendees().size());
		
		Map<String, Attendee> atts = new HashMap<String, Attendee>();
		for(Attendee a : obmEvent.getAttendees()){
			atts.put(a.getEmail(), a);
		}
		
		Attendee phoneOwner = atts.get(bs.getLoginAtDomain());
		Assert.assertNotNull(phoneOwner);
		Assert.assertEquals(bs.getLoginAtDomain(), phoneOwner.getEmail());
		Assert.assertEquals(ParticipationRole.REQ, phoneOwner.getRequired());
		Assert.assertEquals(ParticipationState.ACCEPTED, phoneOwner.getState());
		Assert.assertTrue(phoneOwner.isOrganizer());
		
		Attendee opushAtt = atts.get("email@domain.com");
		Assert.assertNotNull(opushAtt);
		Assert.assertEquals("displayName", opushAtt.getDisplayName());
		Assert.assertEquals("email@domain.com", opushAtt.getEmail());
		Assert.assertEquals(ParticipationRole.REQ, opushAtt.getRequired());
		Assert.assertEquals(ParticipationState.ACCEPTED, opushAtt.getState());
		Assert.assertFalse(opushAtt.isOrganizer());
	}
	
	@Test
	public void testConvertSequence(){
		Event e = getFakeEvent();
		e.setSequence(30);
		IApplicationData ad = converter.convert(getFakeBackendSession(), e);
		Assert.assertTrue(ad instanceof MSEvent);
		MSEvent mse = (MSEvent) ad;
		Assert.assertEquals(30, mse.getObmSequence().intValue());
	}
	
	@Test
	public void testConvertObmSequence(){
		MSEvent e = getFakeMSEvent();
		e.setObmSequence(30);
		Event event = converter.convertAsExternal(getFakeBackendSession(), e);
		Assert.assertEquals(30, event.getSequence());
	}
	
}
