package org.obm.push.backend.obm22.mail;

import java.util.Calendar;
import java.util.Collection;
import java.util.Comparator;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import javax.transaction.UserTransaction;

import org.easymock.EasyMock;
import org.easymock.LogicalOperator;
import org.junit.Assert;
import org.junit.Before;
import org.junit.Test;
import org.minig.imap.FastFetch;
import org.minig.imap.Flag;
import org.minig.imap.SearchQuery;
import org.minig.imap.StoreClient;
import org.minig.obm.pool.OBMPoolActivator;
import org.obm.push.backend.BackendSession;
import org.obm.push.backend.EmailCache;
import org.obm.push.backend.FilterType;
import org.obm.push.state.SyncState;
import org.obm.push.store.ISyncStorage;

import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSet.Builder;

public class EmailSyncCacheTest {

	private ISyncStorage storage;
	private StoreClient imapStore;
	private EmailSyncCache cache;

	@Before
	public void before() {
		storage = EasyMock.createMock(ISyncStorage.class);
		imapStore = EasyMock.createMock(StoreClient.class);
		cache = new EmailSyncCache(storage, getMockOBMPoolActivator());
	}
	
	private MailChanges doSync(SyncState state, Set<FastFetch> mailInImap,
			Integer imapCallExpected, Set<EmailCache> mailSynced, Integer bdCallExpected,
			Collection<Long> removed, Collection<EmailCache> updated)
			throws Throwable {
		return doSync(state, mailInImap, imapCallExpected, mailSynced,new HashSet<EmailCache>(),
				new HashSet<Long>(), bdCallExpected, FilterType.ALL_ITEMS, removed,
				updated);
	}
	
	private MailChanges doSync(SyncState state, Set<FastFetch> mailInImap,
			Integer imapCallExpected, Set<EmailCache> mailSynced, Integer bdCallExpected,
			FilterType filter,
			Collection<Long> removed, Collection<EmailCache> updated)
			throws Throwable {
		return doSync(state, mailInImap, imapCallExpected, mailSynced,new HashSet<EmailCache>(),
				new HashSet<Long>(), bdCallExpected, filter, removed,
				updated);
	}

	private MailChanges doSync(SyncState state, Set<FastFetch> mailInImap,
			Integer imapCallExpected, Set<EmailCache> mailSynced,Set<EmailCache> mailUpdated,
			Set<Long> mailDeleted, Integer bdCallExpected,
			Collection<Long> removed, Collection<EmailCache> updated)
			throws Throwable {
		return doSync(state, mailInImap, imapCallExpected, mailSynced,mailUpdated,
				mailDeleted, bdCallExpected, FilterType.ALL_ITEMS, removed,
				updated);
	}

	@SuppressWarnings("unchecked")
	private MailChanges doSync(SyncState state, 
			Set<FastFetch> mailInImap, Integer imapCallExpected, 
			Set<EmailCache> mailSynced,	Set<EmailCache> mailUpdated, Set<Long> mailDeleted, Integer bdCallExpected, 
			FilterType filter,
			Collection<Long> removed, Collection<EmailCache> updated)
			throws Throwable {
		EasyMock.reset(storage);

		if (bdCallExpected > 0) {
			EasyMock.expect(
					storage.getSyncedMail(getFakeDevId(), getFakeCollectionId()))
					.andReturn(mailSynced).times(bdCallExpected);
			EasyMock.expect(
					storage.getDeletedMail(getFakeDevId(),
							getFakeCollectionId(), state.getLastSync()))
					.andReturn(mailDeleted).times(bdCallExpected);
			EasyMock.expect(
					storage.getUpdatedMail(getFakeDevId(),
							getFakeCollectionId(), state.getLastSync()))
					.andReturn(mailUpdated).times(bdCallExpected);
			
			if (removed.size() > 0) {
				storage.removeMessages(EasyMock.eq(getFakeDevId()),
						EasyMock.eq(getFakeCollectionId()),
						EasyMock.eq(state.getLastSync()),
						compareCollections(removed));
				EasyMock.expectLastCall().times(1);
			}
			if (updated.size() > 0) {
				storage.addMessages(EasyMock.eq(getFakeDevId()),
						EasyMock.eq(getFakeCollectionId()),EasyMock.eq(state.getLastSync()),
						compareCollections(updated));
				EasyMock.expectLastCall().times(1);
			}
		}
		EasyMock.replay(storage);

		EasyMock.reset(imapStore);
		if (imapCallExpected > 0) {
			EasyMock.expect(
					imapStore.uidSearch(EasyMock.anyObject(SearchQuery.class)))
					.andReturn(new HashSet<Long>()).times(imapCallExpected);
			EasyMock.expect(
					imapStore.uidFetchFast(EasyMock.anyObject(Collection.class)))
					.andReturn(mailInImap).times(imapCallExpected);
		}
		EasyMock.replay(imapStore);

		MailChanges mc = cache.getSync(imapStore, getFakeDevId(),
				getFakeBackendSession(), state, getFakeCollectionId(), filter);

		EasyMock.verify(storage);
		EasyMock.verify(imapStore);

		return mc;
	}

	@Test
	public void testEmptyFirstSync() throws Throwable {
		MailChanges changes = doSync(getMockSyncState(new Date(0), "key1"),
				new HashSet<FastFetch>(), 1, new HashSet<EmailCache>(), 1, getDeletedCollection(),
				getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());
	}

	@Test
	public void testTwoSyncWithSameSyncKey() throws Throwable {
		SyncState state = getMockSyncState(new Date(0), "key1");
		MailChanges changes = doSync(state, new HashSet<FastFetch>(), 1,
				new HashSet<EmailCache>(), 1,
				getDeletedCollection(), getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		changes = doSync(state, new HashSet<FastFetch>(), 0,
				new HashSet<EmailCache>(), 0,
				getDeletedCollection(), getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());
	}

	@Test
	public void testUpdatedOnFirstSync() throws Throwable {
		SyncState state = getMockSyncState(new Date(0), "key1");
		FastFetch f1 = getFakeFastFetch(1L, new Date(), false);
		FastFetch f2 = getFakeFastFetch(2L, new Date(), false);
		FastFetch f3 = getFakeFastFetch(3L, new Date(), false);

		MailChanges changes = doSync(state, ImmutableSet.of(f1, f2, f3), 1,
				new HashSet<EmailCache>(), 1,
				getDeletedCollection(), getUpdatedCollection(f1, f2, f3));
		Assert.assertEquals(3, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());
	}

	@Test
	public void testUpdatedOnFirstSyncAndRemoveOnOtherSync() throws Throwable {

		// 3 mail on imap and 0 synced on phone -> 3 up
		FastFetch f1 = getFakeFastFetch(1L, new Date(), false);
		FastFetch f2 = getFakeFastFetch(2L, new Date(), false);
		FastFetch f3 = getFakeFastFetch(3L, new Date(), false);
		MailChanges changes = doSync(getMockSyncState(new Date(0), "key1"),
				ImmutableSet.of(f1, f2, f3), 1, new HashSet<EmailCache>(),
				1, getDeletedCollection(),
				getUpdatedCollection(f1, f2, f3));
		Assert.assertEquals(3, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		// 1 on imap and 3 synced on phone ->2 delete
		EmailCache ec1 = getFakeEmailCache(1L, false);
		EmailCache ec2 = getFakeEmailCache(2L, false);
		EmailCache ec3 = getFakeEmailCache(3L, false);
		changes = doSync(getMockSyncState(new Date(), "key2"),
				ImmutableSet.of(f2), 1, ImmutableSet.of(ec1, ec2, ec3),
				1, getDeletedCollection(f3, f1),
				getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(2, changes.getRemoved().size());

		// 0 on imap and 1 synced on phone ->1 delete
		changes = doSync(getMockSyncState(new Date(), "key3"),
				new HashSet<FastFetch>(), 1, ImmutableSet.of(ec2),
				1, getDeletedCollection(f2),
				getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(1, changes.getRemoved().size());
	}

	@Test
	public void testUpdatedOnFirstSyncAndRemoveUpdateOnOtherSync()
			throws Throwable {
		// 3 mail on imap and 0 synced on phone -> 3 up
		FastFetch f1 = getFakeFastFetch(1L, new Date(), false);
		FastFetch f2 = getFakeFastFetch(2L, new Date(), false);
		FastFetch f3 = getFakeFastFetch(3L, new Date(), false);
		MailChanges changes = doSync(getMockSyncState(new Date(0), "key1"),
				ImmutableSet.of(f1, f2, f3), 1, new HashSet<EmailCache>(),
				1, getDeletedCollection(),
				getUpdatedCollection(f1, f2, f3));
		Assert.assertEquals(3, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		// 1 old and two new on imap, 3 synced on phone ->2 delete and 2 update
		FastFetch f4 = getFakeFastFetch(4L, new Date(), false);
		FastFetch f5 = getFakeFastFetch(5L, new Date(), false);
		Set<FastFetch> imap = ImmutableSet.of(f2, f4, f5);
		EmailCache ec1 = getFakeEmailCache(1L, false);
		EmailCache ec2 = getFakeEmailCache(2L, false);
		EmailCache ec3 = getFakeEmailCache(3L, false);
		Set<EmailCache> cache = ImmutableSet.of(ec1, ec2, ec3);

		changes = doSync(getMockSyncState(new Date(), "key2"), imap, 1, cache,
				1, getDeletedCollection(f1, f3),
				getUpdatedCollection(f4, f5));
		Assert.assertEquals(2, changes.getUpdated().size());
		Assert.assertEquals(2, changes.getRemoved().size());

		// 0 on imap and 1 synced on phone ->1 delete
		changes = doSync(getMockSyncState(new Date(), "key3"),
				new HashSet<FastFetch>(), 1, ImmutableSet.of(ec2),
				1, getDeletedCollection(f2),
				getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(1, changes.getRemoved().size());
	}

	@Test
	public void testFilter() throws Throwable {
		Date now = new Date();
		Calendar cal = Calendar.getInstance();
		cal.setTime(now);
		cal.set(Calendar.DAY_OF_YEAR, cal.get(Calendar.DAY_OF_YEAR) - 1);
		Date oneDayBefore = cal.getTime();
		cal.set(Calendar.DAY_OF_YEAR, cal.get(Calendar.DAY_OF_YEAR) - 9);
		Date tenDayBefore = cal.getTime();
		// 3 mail on imap and 0 synced on phone, filter -> 2 up
		FastFetch oldMail = getFakeFastFetch(1L, tenDayBefore, false);
		FastFetch f2 = getFakeFastFetch(2L, now, false);
		FastFetch f3 = getFakeFastFetch(3L, now, false);
		MailChanges changes = doSync(getMockSyncState(oneDayBefore, "key1"),
				ImmutableSet.of(oldMail, f2, f3), 1, new HashSet<EmailCache>(),
				1, FilterType.THREE_DAYS_BACK,
				getDeletedCollection(), getUpdatedCollection(f2, f3));
		Assert.assertEquals(2, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());
	}

	@Test
	public void testSyncReadUnread() throws Throwable {
		Date now = new Date();
		FastFetch f1 = getFakeFastFetch(1L, now, false);
		FastFetch f2 = getFakeFastFetch(2L, now, false);
		FastFetch f3 = getFakeFastFetch(3L, now, false);
		MailChanges changes = doSync(getMockSyncState(new Date(0), "key1"),
				ImmutableSet.of(f1, f2, f3), 1, new HashSet<EmailCache>(),
				1, getDeletedCollection(),
				getUpdatedCollection(f1, f2, f3));
		Assert.assertEquals(3, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		f1 = getFakeFastFetch(1L, now, true);
		EmailCache ec1 = getFakeEmailCache(1L, false);
		EmailCache ec2 = getFakeEmailCache(2L, false);
		EmailCache ec3 = getFakeEmailCache(3L, false);
		changes = doSync(getMockSyncState(new Date(), "key2"),
				ImmutableSet.of(f1, f2, f3), 1, ImmutableSet.of(ec1, ec2, ec3),
				1, new HashSet<Long>(),
				getUpdatedCollection(f1));
		Assert.assertTrue(changes.getUpdated().contains(new Long(1)));
		Assert.assertEquals(0, changes.getRemoved().size());
	}

	@Test
	public void testAll() throws Throwable {
		Date now = new Date();
		Calendar cal = Calendar.getInstance();
		cal.setTime(now);

		cal.set(Calendar.DAY_OF_YEAR, cal.get(Calendar.DAY_OF_YEAR) - 1);
		Date twoDayBefore = cal.getTime();
		cal.set(Calendar.DAY_OF_YEAR, cal.get(Calendar.DAY_OF_YEAR) - 2);
		Date fourDayBefore = cal.getTime();
		cal.set(Calendar.DAY_OF_YEAR, cal.get(Calendar.DAY_OF_YEAR) - 1);
		Date fiveDayBefore = cal.getTime();

		// 3 mail on imap and 0 synced on phone -> 3 up
		FastFetch f1 = getFakeFastFetch(1L, fourDayBefore, false);
		FastFetch f2 = getFakeFastFetch(2L, twoDayBefore, false);
		FastFetch f3 = getFakeFastFetch(3L, twoDayBefore, false);
		FastFetch f4 = getFakeFastFetch(4L, twoDayBefore, false);
		MailChanges changes = doSync(getMockSyncState(fiveDayBefore, "key1"),
				ImmutableSet.of(f1, f2, f3, f4), 1, new HashSet<EmailCache>(),
				1, FilterType.THREE_DAYS_BACK,
				getDeletedCollection(), getUpdatedCollection(f1, f2, f3, f4));
		Assert.assertEquals(4, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		// f1 -> outOfWindow; f2 -> deleted; f3 ->read; f4 -> no change; f5 and
		// f6 up =>2 delete and 3 update
		f3 = getFakeFastFetch(3L, twoDayBefore, true);
		FastFetch f5 = getFakeFastFetch(5L, new Date(), false);
		FastFetch f6 = getFakeFastFetch(6L, new Date(), false);
		Set<FastFetch> imap = ImmutableSet.of(f3, f4, f5, f6);
		EmailCache ec1 = getFakeEmailCache(1L, false);
		EmailCache ec2 = getFakeEmailCache(2L, false);
		EmailCache ec3 = getFakeEmailCache(3L, false);
		EmailCache ec4 = getFakeEmailCache(4L, false);
		Set<EmailCache> cache = ImmutableSet.of(ec1, ec2, ec3, ec4);
		changes = doSync(getMockSyncState(new Date(), "key2"), imap, 1, cache,
				1, getDeletedCollection(f1, f2),
				getUpdatedCollection(f3, f5, f6));
		Assert.assertEquals(2, changes.getRemoved().size());
		Assert.assertEquals(3, changes.getUpdated().size());

	}

	@Test
	public void testLastSyncWithDeletion() throws Throwable {

		MailChanges changes = doSync(getMockSyncState(new Date(0), "key2"),
				new HashSet<FastFetch>(), 1, new HashSet<EmailCache>(),
				new HashSet<EmailCache>(), ImmutableSet.of(10L, 11L, 12L), 1, getDeletedCollection(),
				new HashSet<EmailCache>());

		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(3, changes.getRemoved().size());

	}

	@Test
	public void testMultiSync() throws Throwable {
		// 3 mail on imap and 0 synced on phone -> 3 up
		FastFetch f1 = getFakeFastFetch(1L, new Date(), false);
		FastFetch f2 = getFakeFastFetch(2L, new Date(), false);
		FastFetch f3 = getFakeFastFetch(3L, new Date(), false);
		MailChanges changes = doSync(getMockSyncState(new Date(0), "key1"),
				ImmutableSet.of(f1, f2, f3), 1, new HashSet<EmailCache>(),
				1, getDeletedCollection(),
				getUpdatedCollection(f1, f2, f3));
		Assert.assertEquals(3, changes.getUpdated().size());
		Assert.assertEquals(0, changes.getRemoved().size());

		// 1 old and two new on imap, 3 synced on phone ->2 delete and 2 update
		SyncState syncSate2 = getMockSyncState(new Date(), "key2");
		FastFetch f4 = getFakeFastFetch(4L, new Date(), false);
		FastFetch f5 = getFakeFastFetch(5L, new Date(), false);
		Set<FastFetch> onimap = ImmutableSet.of(f2, f4, f5);
		EmailCache ec1 = getFakeEmailCache(1L, false);
		EmailCache ec2 = getFakeEmailCache(2L, false);
		EmailCache ec3 = getFakeEmailCache(3L, false);
		Set<EmailCache> onCache = ImmutableSet.of(ec1, ec2, ec3);
		Set<EmailCache> onUpdatedCache = ImmutableSet.of();
		Set<Long> onDeletecache = ImmutableSet.of();
		changes = doSync(syncSate2, onimap, 1, onCache, onUpdatedCache, onDeletecache, 1,
				getDeletedCollection(f1, f3), getUpdatedCollection(f4, f5));
		Assert.assertEquals(2, changes.getUpdated().size());
		Assert.assertEquals(2, changes.getRemoved().size());

		EmailCache ec4 = getFakeEmailCache(4L, false);
		EmailCache ec5 = getFakeEmailCache(5L, false);

		onimap = ImmutableSet.of(f4, f5);
		onCache = ImmutableSet.of(ec2, ec4, ec5);
		onDeletecache = ImmutableSet.of();
		// 0 on imap and 1 synced on phone ->1 delete
		changes = doSync(getMockSyncState(new Date(), "key3"), onimap, 1,
				onCache, onUpdatedCache, onDeletecache, 1, getDeletedCollection(f2),
				getUpdatedCollection());
		Assert.assertEquals(0, changes.getUpdated().size());
		Assert.assertEquals(1, changes.getRemoved().size());

		// replay sync 2
		onCache = ImmutableSet.of(ec4, ec5);
		onUpdatedCache = ImmutableSet.of(ec4, ec5);
		onDeletecache = ImmutableSet.of(f1.getUid(), f2.getUid(), f3.getUid());
		// 0 on imap and 1 synced on phone ->1 delete
		changes = doSync(syncSate2, onimap, 1, onCache, onUpdatedCache, onDeletecache, 1,
				new HashSet<Long>(), getUpdatedCollection());
		Assert.assertEquals(2, changes.getUpdated().size());
		Assert.assertEquals(3, changes.getRemoved().size());
	}

	private FastFetch getFakeFastFetch(Long uid, Date internalDate,
			boolean isRead) {
		Builder<Flag> flag = ImmutableSet.builder();
		if (isRead) {
			flag.add(Flag.SEEN);
		}
		return new FastFetch(uid, internalDate, flag.build());
	}

	private EmailCache getFakeEmailCache(Long uid, boolean isRead) {
		return new EmailCache(uid, isRead);
	}

	private BackendSession getFakeBackendSession() {
		return new BackendSession("adrien@test.tlse.lng", "test", "devID1",
				"devType", "test");
	}

	private Integer getFakeDevId() {
		return 10;
	}

	private Integer getFakeCollectionId() {
		return 100;
	}

	private SyncState getMockSyncState(Date date, String syncKey) {
		SyncState state = EasyMock.createMock(SyncState.class);
		EasyMock.expect(state.getLastSync()).andReturn(date).anyTimes();
		EasyMock.expect(state.getKey()).andReturn(syncKey).anyTimes();
		EasyMock.replay(state);
		return state;
	}

	private OBMPoolActivator getMockOBMPoolActivator() {
		UserTransaction transaction = EasyMock
				.createStrictMock(UserTransaction.class);
		OBMPoolActivator activator = EasyMock
				.createMock(OBMPoolActivator.class);
		EasyMock.expect(activator.getUserTransaction()).andReturn(transaction)
				.anyTimes();
		EasyMock.replay(activator);
		return activator;
	}

	private Collection<EmailCache> getUpdatedCollection() {
		return ImmutableSet.of();
	}

	private Collection<Long> getDeletedCollection() {
		return ImmutableSet.of();
	}

	private Collection<EmailCache> getUpdatedCollection(FastFetch... emails) {
		Builder<EmailCache> b = ImmutableSet.builder();
		for (FastFetch cache : emails) {
			b.add(new EmailCache(cache.getUid(), cache.isRead()));
		}
		return b.build();
	}

	private Collection<Long> getDeletedCollection(FastFetch... fs) {
		Builder<Long> b = ImmutableSet.builder();
		for (FastFetch f : fs) {
			b.add(f.getUid());
		}
		return b.build();
	}

	private static <T, C extends Collection<T>> C compareCollections(
			C collection) {
		return EasyMock.cmp(collection, new Comparator<C>() {
			public int compare(C o1, C o2) {
				if (o1.size() != o2.size()) {
					return -1;
				}
				for (T t : o1) {
					if (!o2.contains(t)) {
						return -1;
					}
				}
				return 0;
			}
		}, LogicalOperator.EQUAL);
	}
}
