package org.obm.push.backend.obm22.mail;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.InetAddress;
import java.util.HashSet;

import org.columba.ristretto.message.Address;
import org.columba.ristretto.smtp.SMTPException;
import org.columba.ristretto.smtp.SMTPProtocol;
import org.easymock.EasyMock;
import org.junit.Before;
import org.junit.Test;
import org.obm.push.backend.BackendSession;
import org.obm.push.backend.obm22.mail.smtp.SmtpLocator;
import org.obm.push.backend.obm22.mail.smtp.SmtpSender;
import org.obm.push.exception.SendEmailException;
import org.obm.push.exception.SmtpInvalidRcptException;
import org.obm.push.exception.SmtpLocatorException;
import org.obm.push.exception.SmtpServiceNotAvailableException;

import com.google.common.collect.ImmutableSet;

public class SmtpSenderTest {
	private SMTPProtocol smtpProtocol;
	private SmtpSender sender;

	@Before
	public void before() throws SmtpLocatorException {
		smtpProtocol = EasyMock.createMock(SMTPProtocol.class);
		SmtpLocator smtpLocator = EasyMock.createMock(SmtpLocator.class);

		EasyMock.expect(
				smtpLocator.getSmtpClient(EasyMock
						.anyObject(BackendSession.class)))
				.andReturn(smtpProtocol).anyTimes();

		EasyMock.replay(smtpLocator);

		sender = new SmtpSender(smtpLocator);
	}

	@Test
	public void ok() throws IOException, SMTPException, SendEmailException, SmtpInvalidRcptException{
		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().times(3);

		smtpProtocol.data(EasyMock.anyObject(InputStream.class));
		EasyMock.expectLastCall().once();

		smtpProtocol.quit();
		EasyMock.expectLastCall().once();

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SmtpServiceNotAvailableException.class)
	public void ioExceptionWhileOpenPort() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {
		EasyMock.expect(smtpProtocol.openPort()).andThrow(new IOException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SendEmailException.class)
	public void smtpExceptionWhileOpenPort() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {
		EasyMock.expect(smtpProtocol.openPort()).andThrow(new SMTPException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SmtpServiceNotAvailableException.class)
	public void ioExceptionWhileEhlo() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {
		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andThrow(new IOException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SendEmailException.class)
	public void smtpExceptionWhileEhlo() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {
		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andThrow(new SMTPException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SmtpServiceNotAvailableException.class)
	public void ioExceptionWhileMail() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().andThrow(new IOException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SendEmailException.class)
	public void smtpExceptionWhileMail() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().andThrow(new SMTPException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SmtpServiceNotAvailableException.class)
	public void ioExceptionWhileRcpt() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().andThrow(new IOException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SendEmailException.class)
	public void smtpExceptionWhileRcpt() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().andThrow(new SMTPException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));
		
		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SmtpServiceNotAvailableException.class)
	public void ioExceptionWhileData() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().times(3);

		smtpProtocol.data(EasyMock.anyObject(InputStream.class));
		EasyMock.expectLastCall().andThrow(new IOException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				new HashSet<Address>(), new HashSet<Address>(),
				new HashSet<Address>(),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test(expected = SendEmailException.class)
	public void smtpExceptionWhileData() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().times(3);

		smtpProtocol.data(EasyMock.anyObject(InputStream.class));
		EasyMock.expectLastCall().andThrow(new IOException());

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andThrow(new SMTPException());

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}

	@Test
	public void smtpExceptionWhileQuit() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(getFakeSender());
		EasyMock.expectLastCall().once();

		smtpProtocol.rcpt(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().times(3);;

		smtpProtocol.data(EasyMock.anyObject(InputStream.class));
		EasyMock.expectLastCall().once();

		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), getFakeSender(),
				ImmutableSet.of(new Address("to1@t.fr")), ImmutableSet.of(new Address("cc2@t.fr")),
				ImmutableSet.of(new Address("bcc3@t.fr")),
				new ByteArrayInputStream(new String().getBytes()));

		EasyMock.verify(smtpProtocol);
	}
	
	@Test
	public void testWithNullParam() throws IOException, SMTPException,
			SendEmailException, SmtpInvalidRcptException {

		EasyMock.expect(smtpProtocol.openPort()).andReturn("Port").once();

		EasyMock.expect(smtpProtocol.ehlo(InetAddress.getLocalHost()))
				.andReturn(new String[0]).once();

		smtpProtocol.mail(EasyMock.anyObject(Address.class));
		EasyMock.expectLastCall().once();

		smtpProtocol.data(EasyMock.anyObject(InputStream.class));
		EasyMock.expectLastCall().once();

		smtpProtocol.quit();
		EasyMock.expectLastCall().once();
		
		EasyMock.replay(smtpProtocol);

		sender.sendEmail(getFakeBackendSession(), null, null, null, null, null);

		EasyMock.verify(smtpProtocol);
	}

	private BackendSession getFakeBackendSession() {
		return new BackendSession("adrien@test.tlse.lng", "test", "devID1",
				"devType", "test");
	}
	
	private Address getFakeSender() {
		return new Address("addr@t.fr");
	}
}
